# -*- coding: utf-8 -*-
import inspect
from functools import wraps
from string import Formatter
from pydantic.main import BaseModel
from easi_py_common.client.rest_client import RestClient, DST_MODE_SERVICE

service_endpoint = {}

HTTP_METHOD_GET = 'GET'
HTTP_METHOD_POST = 'POST'
HTTP_METHOD_PUT = 'PUT'
HTTP_METHOD_DELETE = 'DELETE'


def service_client(url, timeout=(5, 5), http_adapter=None, before_request=None,
                   dst_mode=DST_MODE_SERVICE):
    def wrapper(f):
        key = '{}.{}'.format(f.__module__, f.__name__)

        service_endpoint[key] = RestClient(
            url, timeout=timeout, http_adapter=http_adapter, before_request=before_request,
            dst_mode=dst_mode)

        @wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapped

    return wrapper


def service_mapping(uri, method, timeout=None, data_by_alias: bool = True, data_key='data'):
    def wrapper(f):
        full_arg_spec = inspect.getfullargspec(f)
        is_method = full_arg_spec.args and full_arg_spec.args[0] == 'self'
        if not is_method:
            raise Exception('{} is invalid service_mapping', f.__name__)

        dst = full_arg_spec.annotations.get('return')
        has_format_names = [fn for _, fn, _, _ in Formatter().parse(uri) if fn is not None]
        has_format_names.append('self')  # 参数列表过滤self

        def __get_request_params(named_args):
            request_named_args = {k: v for k, v in named_args.items() if k not in has_format_names}
            data = None
            if method == HTTP_METHOD_POST or method == HTTP_METHOD_PUT:
                if data_key in request_named_args:
                    data = request_named_args.pop(data_key)
                if data is None:
                    data = {}

            return request_named_args, data

        def __get_function_args(args, kwargs):
            named_args = inspect.getcallargs(f, *args, **kwargs)
            named_args.update(named_args.pop(full_arg_spec.varkw, {}))
            if ((method == HTTP_METHOD_GET or method == HTTP_METHOD_DELETE)
                    and data_key in named_args and isinstance(named_args[data_key], BaseModel)):
                params_data = named_args.pop(data_key)
                named_args.update(params_data.dict(by_alias=data_by_alias))
            return named_args

        @wraps(f)
        def wrapped(*args, **kwargs):
            name = '{}.{}'.format(args[0].__class__.__module__, args[0].__class__.__name__)
            if name not in service_endpoint:
                raise Exception('{}.{} is not have config endpoint', f.__module__, f.__name__)

            named_args = __get_function_args(args, kwargs)
            _uri = uri.format(**named_args)
            params, data = __get_request_params(named_args)

            client = service_endpoint[name]
            if method == HTTP_METHOD_GET:
                return client.get(_uri, params=params, timeout=timeout, dst=dst)
            elif method == HTTP_METHOD_POST:
                return client.post(_uri, json_data=data, params=params, timeout=timeout, dst=dst,
                                   request_by_alias=data_by_alias)
            elif method == HTTP_METHOD_PUT:
                return client.put(_uri, json_data=data, params=params, timeout=timeout, dst=dst,
                                  request_by_alias=data_by_alias)
            elif method == HTTP_METHOD_DELETE:
                return client.delete(_uri, params=params, timeout=timeout, dst=dst)
            else:
                raise Exception('{}.{} is invalid method', f.__module__, f.__name__)

        return wrapped

    return wrapper


def service_mapping_get(uri, timeout=None, data_by_alias: bool = True, data_key='data'):
    return service_mapping(uri, method=HTTP_METHOD_GET, timeout=timeout,
                           data_by_alias=data_by_alias, data_key=data_key)


def service_mapping_post(uri, timeout=None, data_by_alias: bool = True, data_key='data'):
    return service_mapping(uri, method=HTTP_METHOD_POST, timeout=timeout,
                           data_by_alias=data_by_alias, data_key=data_key)


def service_mapping_put(uri, timeout=None, request_by_alias: bool = True, data_key='data'):
    return service_mapping(uri, method=HTTP_METHOD_PUT, timeout=timeout,
                           data_by_alias=request_by_alias, data_key=data_key)


def service_mapping_delete(uri, timeout=None, request_by_alias: bool = True, data_key='data'):
    return service_mapping(uri, method=HTTP_METHOD_DELETE, timeout=timeout,
                           data_by_alias=request_by_alias, data_key=data_key)



import time
from datetime import datetime

DATE_TIME_FORMAT = '%Y.%m.%d %H:%M:%S'


def format_timestamp(timestamp, formats=DATE_TIME_FORMAT, is_millisecond=True):
    if is_millisecond:
        timestamp = round(timestamp / 1000.0, 2)
    date_array = datetime.utcfromtimestamp(timestamp)
    return date_array.strftime(formats)


def get_utc_timestamp(is_millisecond=True):
    timestamp = time.time()
    return int(timestamp * 1000) if is_millisecond else timestamp


def is_empty(string):
    if string is None:
        return True

    if type(string) == str:
        return len(string.strip()) == 0
    else:
        return False


def is_not_empty(string):
    return not is_empty(string)



import os

from inspect import isfunction
from typing import Optional, List

import yaml
from addict import Dict
from aenum import Enum, NoAlias

from easi_py_common.core.error import ServiceException
from easi_py_common.core.sql_alchemy import db
from flask import current_app, request, Flask

LANG_ZH = 'zh'

LANG_CN = 'cn'
LANG_EN = 'en'
LANG_JA = 'ja'
LANG_MS = 'ms'

LANG_LIST = []


def verify_language(lang) -> bool:
    if not lang:
        return False
    return lang in LANG_LIST


def get_header_language() -> str:
    lang = None

    if current_app.header_language_function is not None:
        func_lang = current_app.header_language_function()
        if verify_language(func_lang):
            lang = func_lang

    if not lang:
        header_lang = request.headers.get(current_app.header_language_name)
        if verify_language(header_lang):
            lang = header_lang

    if not lang:
        lang = current_app.i18n_language_default or LANG_EN

    if lang == LANG_ZH:
        lang = LANG_CN

    lang = current_app.i18n_language_mapping.get(lang) or lang
    return lang


def verify_auto_language(lang: Optional[str]) -> str:
    if verify_language(lang):
        return lang
    return get_header_language()


def verify_default_language(lang: str, default_lang: str = LANG_EN) -> str:
    if verify_language(lang):
        return lang
    if verify_language(default_lang):
        return default_lang
    return LANG_EN


def trans_data(data_cn, data_en, lang=None):
    lang = verify_auto_language(lang)
    if lang in (LANG_CN, LANG_ZH):
        return data_cn

    return handle_lang_en_val(data_en, data_cn)


def trans(data, data_key, lang=None):
    lang = verify_auto_language(lang)

    if not isinstance(data, (dict, db.Model)):
        return ''

    if isinstance(data, db.Model):
        data = data.__dict__

    return trans_data(data.get(data_key), data.get('{}_en'.format(data_key)), lang=lang)


def tt(translate_key: str, lang: Optional[str] = None) -> Dict:
    """
    1. 指定返回参数值。两个参数都不能为空,语言为空则返回中文内容

    :param translate_key: 参数。解析 en.yml cn.yml 语言配置文件, 只需要传入 translate_key, 自动返回该配置文件下该 key 的字典
    :param lang: 指定的具体语言
    :return:
    """
    return current_app.i18n.get(verify_auto_language(lang), Dict()).get(translate_key, Dict())


def t(translate_key=None, lang=None, data=None, data_key=None, data_cn=None, data_en=None):
    """
    1. 指定返回参数值。两个参数都不能为空,语言为空则返回中文内容

    :param translate_key: 参数。解析 en.yml cn.yml 语言配置文件, 只需要传入 translate_key, 自动返回该配置文件下该 key 的字典
    :param lang: 指定的具体语言
    :param data:
    :param data_key:
    :param data_cn:
    :param data_en:
    :return:
    """
    lang = verify_auto_language(lang)
    if translate_key:
        return tt(translate_key, lang=lang)

    if data_cn is not None or data_en is not None:
        return trans_data(data_cn, data_en, lang=lang)

    return trans(data, data_key, lang=lang)


def is_empty(string):
    if string is None:
        return True

    if type(string) == str:
        return len(string.strip()) == 0
    else:
        return False


def handle_lang_en_val(en_val, zh_val):
    """
    如果英文为空会返回中文
    :param en_val:
    :param zh_val:
    :return:
    """
    return zh_val if is_empty(en_val) else en_val


class BaseErrorCode(Enum, settings=NoAlias):

    def __get_code(self):
        code = 400
        def_msg = ''
        if isinstance(self.value, int):
            code = self.value
        elif isinstance(self.value, tuple):
            code, def_msg = self.value
        return code, def_msg

    def __get_code_message(self):
        code, def_msg = self.__get_code()
        return code, t(self.error_code_prefix()).get(self.name) or def_msg

    def error_code_prefix(self):
        return 'error_code'

    def build_exception(self):

        code, msg = self.__get_code_message()
        return ServiceException(code=code, msg=msg)

    def build_invalid_params_exception(self, msg):
        code, def_msg = self.__get_code()
        return ServiceException(code=code, msg=msg or def_msg)

    def build_params_exception(self, *args):
        code, msg = self.__get_code_message()
        return ServiceException(code=code, msg=msg.format(*args))

    def build_key_params_exception(self, **kwargs):
        code, msg = self.__get_code_message()
        return ServiceException(code=code, msg=msg.format(**kwargs))

    def build_value_error(self):
        code, msg = self.__get_code_message()
        return ValueError(msg)

    def build_params_value_error(self, *args):
        code, msg = self.__get_code_message()
        return ValueError(msg.format(*args))

    def build_key_params_value_error(self, **kwargs):
        code, msg = self.__get_code_message()
        return ValueError(msg.format(**kwargs))


def init_app(app: Flask, lang_list: List[str], header_language_name: str = 'language',
             header_language_function=None, i18n_language_default: str = None, i18n_language_mapping: dict = None):
    if header_language_function and not isfunction(header_language_function):
        raise ServiceException(code=400, msg='header_language_function 不是  function')

    if not lang_list:
        lang_list = [LANG_CN, LANG_EN, LANG_JA, LANG_MS]

    app_path = os.path.dirname(app.root_path)
    lang_map = {}
    for lang in lang_list:
        lang_yml_path = os.path.join(app_path, 'common/i18n/{}.yml'.format(lang))
        with open(lang_yml_path) as lang_yml_file:
            lang_yml = yaml.load(lang_yml_file, Loader=yaml.FullLoader)
            lang_map[lang] = Dict(lang_yml)
            if lang not in LANG_LIST:
                LANG_LIST.append(lang)

    app.i18n = lang_map
    app.header_language_name = header_language_name
    app.header_language_function = header_language_function
    app.i18n_language_default = i18n_language_default if i18n_language_default in lang_list else LANG_EN
    app.i18n_language_mapping = i18n_language_mapping or dict()



# -*- coding: utf-8 -*-

class ConsumerLogConstants:
    PUB_LOG = '{}_pub_log'
    SUB_LOG = '{}_sub_log'

    CONSUMER_LOG_ES_BULK_NUM = 1000
    CONSUMER_LOG_ES_BULK_THREAD_NUM = 8
    CONSUMER_LOG_ES_BULK_BUFFER_SIZE = CONSUMER_LOG_ES_BULK_NUM * CONSUMER_LOG_ES_BULK_THREAD_NUM * 2
    CONSUMER_LOG_MAX_SLEEP_ES_BULK_BUFFER_SIZE = CONSUMER_LOG_ES_BULK_BUFFER_SIZE * 1.5

# 日志
PUB_LOG = 'pub_log'
SUB_LOG = 'sub_log'
PUB_LOG_BULK_SAVE = 'pub_log_bulk_save'
SUB_LOG_BULK_SAVE = 'sub_log_bulk_save'



# -*- coding: utf-8 -*-
import json
import time
import traceback
from datetime import datetime
from inspect import isfunction

from flask import current_app

from easi_py_common.core.redis import redis_store, RedisLock
from easi_py_common.sqs import utils
from easi_py_common.sqs.constants import ConsumerLogConstants


def get_consumer_log_es_bulk_key():
    # 生成锁key
    consumer_log_es_bulk_key_template = 'delivery_consumer_log_es_bulk_key_{}'
    consumer_log_es_bulk_key_list = []
    for index in range(ConsumerLogConstants.CONSUMER_LOG_ES_BULK_THREAD_NUM):
        consumer_log_es_bulk_key_list.append(consumer_log_es_bulk_key_template.format(index))

    # 生成锁key 的值
    consumer_log_es_bulk_key_value_list = redis_store.mget(consumer_log_es_bulk_key_list)

    # 进行判断是否有可用锁
    consumer_log_es_bulk_key = None
    for index, value in enumerate(consumer_log_es_bulk_key_value_list):
        if not value:
            consumer_log_es_bulk_key = consumer_log_es_bulk_key_template.format(index)
            break

    if not consumer_log_es_bulk_key:
        return None

    # 进行锁20s
    redis_store.set(consumer_log_es_bulk_key, 1, ex=20)
    return consumer_log_es_bulk_key


class SqsPubLogDTO:
    msg_id = None
    address = None
    url = None
    k = None
    delay_seconds = None
    body = None
    group_id = None
    work_wait_time = None
    send_execute_time = None
    send_time = None
    send_success_time = None
    create_time = None


class SqsSubLogDTO:
    msg_id = None
    address = None
    url = None
    k = None
    body = None
    delay_seconds = None
    status = None
    receive_time = None
    receive_success_time = None
    message_wait_time = None
    work_wait_time = None
    execute_time = None
    create_time = None


def get_log_list(log_list_key, log_app_name):
    with RedisLock(lock_name="{}_consumer_log_es_{}_bulk_num_cat".format(log_app_name, log_list_key), timeout_ms=6000):
        # 获取现在的日志长度
        sub_log_len = redis_store.llen(log_list_key)
        if sub_log_len < ConsumerLogConstants.CONSUMER_LOG_ES_BULK_NUM:
            current_app.logger.info('数量不够')
            return [], ''

        # 获取锁key
        consumer_log_es_bulk_key = get_consumer_log_es_bulk_key()
        if not consumer_log_es_bulk_key:
            current_app.logger.info('没有获取到锁')
            return [], ''

        # 获取保存的日志列表
        pipe = redis_store.pipeline()
        for _ in range(ConsumerLogConstants.CONSUMER_LOG_ES_BULK_NUM):
            pipe.rpop(log_list_key)
        return pipe.execute(), consumer_log_es_bulk_key


class LogSqs:
    buffer_size = 1500
    buffer_cut = 500

    @staticmethod
    def save_producer_log(log_app_name, save_func):
        if not isfunction(save_func):
            return False
        # 获取需要保存的日志列表和锁key
        exec_results_start_time = time.time()
        results, consumer_log_es_bulk_key = get_log_list(ConsumerLogConstants.PUB_LOG.format(log_app_name),
                                                         log_app_name)
        exec_results_end_time = time.time()
        if exec_results_end_time - exec_results_start_time > 1:
            current_app.logger.error(
                'producer_log: get_results:{}'.format(exec_results_end_time - exec_results_start_time))

        if not results:
            return False

        batch_saves_start_time = time.time()
        try:
            batch_saves = []
            for m in results:
                if not m:
                    continue

                msg_ret = json.loads(m)
                msg_body = msg_ret['msg']
                if isinstance(msg_body, dict):
                    try:
                        msg_body = json.dumps(msg_body)
                        msg_body = msg_body[0:2000]
                    except Exception:
                        current_app.logger.error(
                            'producer_log resolve msg body error:{}, '
                            'body:{}'.format(traceback.format_exc(), msg_ret['body'])
                        )

                send_execute_time = -1
                try:
                    send_execute_time_seconds = (
                            datetime.strptime(msg_ret['send_success_time'], '%Y.%m.%d %H:%M:%S.%f')
                            - datetime.strptime(msg_ret['send_time'], '%Y.%m.%d %H:%M:%S.%f')
                    ).total_seconds()
                    send_execute_time = int(send_execute_time_seconds * 1000)
                except Exception:
                    pass

                work_wait_time = -1
                if 'work_wait_time' in msg_ret:
                    work_wait_time = msg_ret['work_wait_time']

                create_time = utils.format_timestamp(utils.get_utc_timestamp(), formats='%Y.%m.%d %H:%M:%S.%f')[:-3]
                send_time = msg_ret['send_time']
                send_success_time = msg_ret['send_success_time']

                pub_log = SqsPubLogDTO()
                pub_log.msg_id = msg_ret['msg_id']
                pub_log.address = msg_ret['address']
                pub_log.url = msg_ret['url']
                pub_log.k = msg_ret['key']
                pub_log.delay_seconds = msg_ret['delay_seconds']
                pub_log.body = msg_body
                pub_log.group_id = msg_ret['group_id']
                pub_log.work_wait_time = work_wait_time
                pub_log.send_execute_time = send_execute_time
                pub_log.send_time = send_time
                pub_log.send_success_time = send_success_time
                pub_log.create_time = create_time

                batch_saves.append(pub_log)

            if batch_saves:
                save_func(batch_saves)
        except Exception as e:
            current_app.logger.error('producer_log batch_saves error:{}'.format(traceback.format_exc()))
            raise e
        finally:
            # current_app.logger.error('producer_log batch_saves3')
            if consumer_log_es_bulk_key:
                redis_store.delete(consumer_log_es_bulk_key)

            batch_saves_total_time = time.time() - batch_saves_start_time
            if batch_saves_total_time > 15:
                current_app.logger.error('producer_log batch_saves_time:{}'.format(batch_saves_total_time))

        return True

    @staticmethod
    def save_consumer_log(log_app_name, save_func):
        if not isfunction(save_func):
            return False
        # 获取需要保存的日志列表和锁key
        exec_results_start_time = time.time()
        results, consumer_log_es_bulk_key = get_log_list(ConsumerLogConstants.SUB_LOG.format(log_app_name),
                                                         log_app_name)
        exec_results_end_time = time.time()
        if exec_results_end_time - exec_results_start_time > 1:
            current_app.logger.error(
                'consumer_log: get_results:{}'.format(exec_results_end_time - exec_results_start_time))

        if not results:
            return False

        batch_saves_start_time = time.time()
        try:
            batch_saves = []
            for m in results:
                if not m:
                    continue

                msg_ret = json.loads(m)
                msg_ret['create_time'] = utils.format_timestamp(utils.get_utc_timestamp(),
                                                                formats='%Y.%m.%d %H:%M:%S.%f')[:-3]

                if 'work_wait_time' not in msg_ret:
                    msg_ret['work_wait_time'] = -1
                if 'message_wait_time' not in msg_ret:
                    msg_ret['message_wait_time'] = -1
                if 'delay_seconds' not in msg_ret:
                    msg_ret['delay_seconds'] = -1

                msg_body = msg_ret['body']
                if utils.is_not_empty(msg_body):
                    try:
                        msg_body_json = json.loads(msg_body)
                        if isinstance(msg_body_json, dict) and 't' in msg_body_json:
                            msg_body_msg = msg_body_json['msg']
                            if utils.is_not_empty(msg_body_msg):
                                msg_body = json.dumps(msg_body_msg)
                            else:
                                msg_body = ''

                        msg_ret['body'] = msg_body[0:2000]
                    except Exception:
                        current_app.logger.error(
                            'consumer_log resolve msg body error:{}, '
                            'body:{}'.format(traceback.format_exc(), msg_ret['body'])
                        )
                msg_ret['execute_time'] = -1
                try:
                    if 'receive_success_time' in msg_ret and msg_ret['receive_success_time']:
                        execute_time = (
                                datetime.strptime(msg_ret['receive_success_time'], '%Y.%m.%d %H:%M:%S.%f')
                                - datetime.strptime(msg_ret['receive_time'], '%Y.%m.%d %H:%M:%S.%f')
                        ).total_seconds()
                        msg_ret['execute_time'] = int(execute_time * 1000)
                    else:
                        msg_ret['receive_success_time'] = msg_ret['receive_time']
                except Exception:
                    pass

                sub_log = SqsSubLogDTO()
                sub_log.msg_id = msg_ret['msg_id']
                sub_log.address = msg_ret['address']
                sub_log.url = msg_ret['url']
                sub_log.k = msg_ret['key']
                sub_log.body = msg_ret['body']
                sub_log.delay_seconds = msg_ret['delay_seconds']
                sub_log.status = msg_ret['status']
                sub_log.receive_time = msg_ret['receive_time']
                sub_log.receive_success_time = msg_ret['receive_success_time']
                sub_log.message_wait_time = msg_ret['message_wait_time']
                sub_log.work_wait_time = msg_ret['work_wait_time']
                sub_log.execute_time = msg_ret['execute_time']
                sub_log.create_time = msg_ret['create_time']

                batch_saves.append(sub_log)

            if batch_saves:
                save_func(batch_saves)

        except Exception as e:
            current_app.logger.error('consumer_log batch_saves error:{}'.format(traceback.format_exc()))
            raise e
        finally:
            # current_app.logger.error('consumer_log batch_saves3')
            if consumer_log_es_bulk_key:
                redis_store.delete(consumer_log_es_bulk_key)

            batch_saves_total_time = time.time() - batch_saves_start_time
            if batch_saves_total_time > 15:
                current_app.logger.error('consumer_log batch_saves_time:{}'.format(batch_saves_total_time))

        return True
